/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007, 2008, 2009 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Tests for the prime field arithmetic module.
 *
 * @version $Id: test_fp.c 261 2010-02-05 00:39:28Z conradoplg $
 * @ingroup test
 */

#include <stdio.h>

#include "relic.h"
#include "relic_test.h"
#include "relic_fp_low.h"


static int memory(void) {
	err_t e;
	int code = STS_ERR;
	fp_t a;

	fp_null(a);

	TRY {
		TEST_BEGIN("memory can be allocated") {
			fp_new(a);
			fp_free(a);
		} TEST_END;
	} CATCH(e) {
		switch (e) {
			case ERR_NO_MEMORY:
				util_print("FATAL ERROR!\n");
				ERROR(end);
				break;
		}
	}
	(void)a;
	code = STS_OK;
  end:
	return code;
}

static int util(void) {
	int bits, code = STS_ERR;
	char str[1000];
	fp_t a, b, c;

	fp_null(a);
	fp_null(b);
	fp_null(c);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("comparison is consistent") {
			fp_rand(a);
			fp_rand(b);
			if (fp_cmp(a, b) != CMP_EQ) {
				if (fp_cmp(a, b) == CMP_GT) {
					TEST_ASSERT(fp_cmp(b, a) == CMP_LT, end);
				} else {
					TEST_ASSERT(fp_cmp(b, a) == CMP_GT, end);
				}
			}
		}
		TEST_END;

		TEST_BEGIN("copy and comparison are consistent") {
			fp_rand(a);
			fp_rand(c);
			fp_rand(b);
			if (fp_cmp(a, c) != CMP_EQ) {
				fp_copy(c, a);
				TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
			}
			if (fp_cmp(b, c) != CMP_EQ) {
				fp_copy(c, b);
				TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
			}
		}
		TEST_END;

		TEST_BEGIN("negation is consistent") {
			fp_rand(a);
			fp_neg(b, a);
			if (fp_cmp(a, b) != CMP_EQ) {
				if (fp_cmp(a, b) == CMP_GT) {
					TEST_ASSERT(fp_cmp(b, a) == CMP_LT, end);
				} else {
					TEST_ASSERT(fp_cmp(b, a) == CMP_GT, end);
				}
			}
			fp_neg(b, b);
			TEST_ASSERT(fp_cmp(a, b) == CMP_EQ, end);
		}
		TEST_END;

		TEST_BEGIN("assignment to zero and comparison are consistent") {
			fp_rand(a);
			fp_zero(c);
			TEST_ASSERT(fp_cmp(a, c) == CMP_GT, end);
			TEST_ASSERT(fp_cmp(c, a) == CMP_LT, end);
		}
		TEST_END;

		TEST_BEGIN("assignment to random and comparison are consistent") {
			fp_rand(a);
			fp_rand(b);
			fp_zero(c);
			TEST_ASSERT(fp_cmp(a, c) == CMP_GT, end);
			TEST_ASSERT(fp_cmp(b, c) == CMP_GT, end);
			if (fp_cmp(a, b) != CMP_EQ) {
				if (fp_cmp(a, b) == CMP_GT) {
					TEST_ASSERT(fp_cmp(b, a) == CMP_LT, end);
				} else {
					TEST_ASSERT(fp_cmp(b, a) == CMP_GT, end);
				}
			}
		}
		TEST_END;

		TEST_BEGIN("assignment to zero and zero test are consistent") {
			fp_zero(a);
			TEST_ASSERT(fp_is_zero(a), end);
		}
		TEST_END;

		bits = 0;
		TEST_BEGIN("bit setting and testing are consistent") {
			fp_zero(a);
			fp_set_bit(a, bits, 1);
			TEST_ASSERT(fp_test_bit(a, bits), end);
			bits = (bits + 1) % FP_BITS;
		}
		TEST_END;

		bits = 0;
		TEST_BEGIN("bit assignment and counting are consistent") {
			fp_zero(a);
			fp_set_bit(a, bits, 1);
			TEST_ASSERT(fp_bits(a) == bits + 1, end);
			bits = (bits + 1) % FP_BITS;
		}
		TEST_END;

		TEST_BEGIN("reading and writing a prime field element are consistent") {
			fp_rand(a);
			fp_write(str, sizeof(str), a, 16);
			fp_read(b, str, sizeof(str), 16);
			TEST_ASSERT(fp_cmp(a, b) == CMP_EQ, end);
		}
		TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

static int addition(void) {
	int code = STS_ERR;
	fp_t a, b, c, d, e;
	dv_t A, B, C, D, E;

	fp_null(a);
	fp_null(b);
	fp_null(c);
	fp_null(d);
	fp_null(e);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);
		fp_new(d);
		fp_new(e);
                dv_new(A);
                dv_new(B);
                dv_new(C);
                dv_new(D);
                dv_new(E);

		TEST_BEGIN("addition is commutative") {
			fp_rand(a);
			fp_rand(b);
			fp_add(d, a, b);
			fp_add(e, b, a);
			TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("addition is associative") {
			fp_rand(a);
			fp_rand(b);
			fp_rand(c);
			fp_add(d, a, b);
			fp_add(d, d, c);
			fp_add(e, b, c);
			fp_add(e, a, e);
			TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("addition has identity") {
			fp_rand(a);
			fp_zero(d);
			fp_add(e, a, d);
			TEST_ASSERT(fp_cmp(e, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("addition has inverse") {
			fp_rand(a);
			fp_neg(d, a);
			fp_add(e, a, d);
			TEST_ASSERT(fp_is_zero(e), end);
		} TEST_END;

                TEST_BEGIN("double-precision addition is commutative") {
		  fp_rand(a);
		  fp_rand(b);
                  fp_rand(c);
                  fp_rand(d);
                  fp_muln_low(A,a,b);
                  fp_muln_low(B,c,d);
		  fp_addd(C, A, B);
		  fp_addd(D, B, A);
                  fp_rdc(c,C);
                  fp_rdc(d,D);
		  TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
                } TEST_END;

                TEST_BEGIN("double-precision addition is associative") {
		  fp_rand(a);
		  fp_rand(b);
		  fp_rand(c);
                  fp_rand(d);
                  fp_muln_low(A,a,b);
                  fp_muln_low(B,c,d);
                  fp_muln_low(C,a,d);
		  fp_addd(D, A, B);
		  fp_addd(D, D, C);
		  fp_addd(E, B, C);
		  fp_addd(E, A, E);
                  fp_rdc(d,D);
                  fp_rdc(e,E);
		  TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
                } TEST_END;

                TEST_BEGIN("double-precision addition has identity") {
                  fp_rand(a);
                  fp_rand(b);
		  fp_zero(d);
                  fp_muln_low(A, a, b);
                  fp_muln_low(D, d, d);
		  fp_addd(E, A, D);
                  fp_rdc(a, A);
                  fp_rdc(e, E);
		  TEST_ASSERT(fp_cmp(e, a) == CMP_EQ, end);
                } TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	fp_free(d);
	fp_free(e);
	dv_free(A);
	dv_free(B);
	dv_free(C);
	dv_free(D);
        dv_free(E);
	return code;
}

static int subtraction(void) {
	int code = STS_ERR;
	fp_t a, b, c, d;
	dv_t A,B,C,D;

	fp_null(a);
	fp_null(b);
	fp_null(c);
	fp_null(d);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);
		fp_new(d);
                dv_new(A);
                dv_new(B);
                dv_new(C);
                dv_new(D);

		TEST_BEGIN("subtraction is anti-commutative") {
			fp_rand(a);
			fp_rand(b);
			fp_sub(c, a, b);
			fp_sub(d, b, a);
			fp_neg(d, d);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		}
		TEST_END;

		TEST_BEGIN("subtraction has identity") {
			fp_rand(a);
			fp_zero(c);
			fp_sub(d, a, c);
			TEST_ASSERT(fp_cmp(d, a) == CMP_EQ, end);
		}
		TEST_END;

		TEST_BEGIN("subtraction has inverse") {
			fp_rand(a);
			fp_sub(c, a, a);
			TEST_ASSERT(fp_is_zero(c), end);
		}
		TEST_END;

                TEST_BEGIN("double-precision subtraction is anti-commutative") {
		  fp_rand(a);
		  fp_rand(b);
                  fp_rand(c);
                  fp_rand(d);
                  fp_muln_low(A, a, b);
                  fp_muln_low(B, c, d);
		  fp_subd(C, A, B);
                  fp_rdc(c, C);
		  fp_subd(D, B, A);
                  fp_rdc(d, D);
		  fp_neg(d, d);
		  TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
                }
                TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	fp_free(d);
        dv_free(A);
        dv_free(B);
        dv_free(C);
        dv_free(D);
	return code;
}

static int multiplication(void) {
	int code = STS_ERR;
	fp_t a, b, c, d, e, f;
        dv_t D;

	fp_null(a);
	fp_null(b);
	fp_null(c);
	fp_null(d);
	fp_null(e);
	fp_null(f);
        dv_null(D);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);
		fp_new(d);
		fp_new(e);
		fp_new(f);
                dv_new(D);

		TEST_BEGIN("multiplication is commutative") {
			fp_rand(a);
			fp_rand(b);
			fp_mul(d, a, b);
			fp_mul(e, b, a);
			TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("multiplication is associative") {
			fp_rand(a);
			fp_rand(b);
			fp_rand(c);
			fp_mul(d, a, b);
			fp_mul(d, d, c);
			fp_mul(e, b, c);
			fp_mul(e, a, e);
			TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("multiplication is distributive") {
			fp_rand(a);
			fp_rand(b);
			fp_rand(c);
			fp_add(d, a, b);
			fp_mul(d, c, d);
			fp_mul(e, c, a);
			fp_mul(f, c, b);
			fp_add(e, e, f);
			TEST_ASSERT(fp_cmp(d, e) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("multiplication has identity") {
			fp_rand(a);
			fp_set_dig(d, 1);
			fp_mul(e, a, d);
			TEST_ASSERT(fp_cmp(e, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("multiplication has zero property") {
			fp_rand(a);
			fp_zero(d);
			fp_mul(e, a, d);
			TEST_ASSERT(fp_is_zero(e), end);
		} TEST_END;

#if FP_MUL == BASIC || !defined(STRIP)
		TEST_BEGIN("basic multiplication is correct") {
			fp_rand(a);
			fp_rand(b);
			fp_mul(c, a, b);
			fp_mul_basic(d, a, b);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		}
		TEST_END;
#endif

#if FP_MUL == INTEG || !defined(STRIP)
		TEST_BEGIN("integrated multiplication is correct") {
			fp_rand(a);
			fp_rand(b);
			fp_mul(c, a, b);
			fp_mul_integ(d, a, b);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		}
		TEST_END;
#endif

#if FP_MUL == COMBA || !defined(STRIP)
		TEST_BEGIN("comba multiplication is correct") {
			fp_rand(a);
			fp_rand(b);
			fp_mul(c, a, b);
			fp_mul_comba(d, a, b);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		}
		TEST_END;
#endif

#if FP_KARAT > 0 || !defined(STRIP)
		TEST_BEGIN("karatsuba multiplication is correct") {
			fp_rand(a);
			fp_rand(b);
			fp_mul(c, a, b);
			fp_mul_karat(d, a, b);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		}
		TEST_END;
#endif

#if FP_MUL == COMBA || !defined(STRIP)
                TEST_BEGIN("unreduced comba multiplication is correct") {
		  fp_rand(a);
		  fp_rand(b);
		  fp_mul(c, a, b);
		  fp_muln_low(D, a, b);
                  fp_rdc(d,D);
		  TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
                }
                TEST_END;
#endif

	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	fp_free(d);
	fp_free(e);
	fp_free(f);
        dv_free(D);
	return code;
}

static int squaring(void) {
	int code = STS_ERR;
	fp_t a, b, c;

	fp_null(a);
	fp_null(b);
	fp_null(c);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("squaring is correct") {
			fp_rand(a);
			fp_mul(b, a, a);
			fp_sqr(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;

#if FP_SQR == BASIC || !defined(STRIP)
		TEST_BEGIN("basic squaring is correct") {
			fp_rand(a);
			fp_sqr(b, a);
			fp_sqr_basic(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;
#endif

#if FP_SQR == INTEG || !defined(STRIP)
		TEST_BEGIN("integrated squaring is correct") {
			fp_rand(a);
			fp_sqr(b, a);
			fp_sqr_integ(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;
#endif

#if FP_SQR == COMBA || !defined(STRIP)
		TEST_BEGIN("comba squaring is correct") {
			fp_rand(a);
			fp_sqr(b, a);
			fp_sqr_comba(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;
#endif

#if FP_KARAT > 0 || !defined(STRIP)
		TEST_BEGIN("karatsuba squaring is correct") {
			fp_rand(a);
			fp_sqr(b, a);
			fp_sqr_karat(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;
#endif
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

static int doubling_halving(void) {
	int code = STS_ERR;
	fp_t a, b, c;

	fp_null(a);
	fp_null(b);
	fp_null(c);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("doubling is consistent") {
			fp_rand(a);
			fp_add(b, a, a);
			fp_dbl(c, a);
			TEST_ASSERT(fp_cmp(b, c) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("halving is consistent") {
			fp_rand(a);
			if (!fp_is_even(a)) {
				a[0] &= MASK(FP_DIGIT - 1) << 1;
			}
			fp_hlv(b, a);
			fp_dbl(c, b);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		}
		TEST_END;
	}
	CATCH_ANY {
		util_print("FATAL ERROR!\n");
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

static int shifting(void) {
	int code = STS_ERR;
	fp_t a, b, c;

	fp_null(a);
	fp_null(b);
	fp_null(c);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("shifting by 1 bit is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			fp_lsh(b, a, 1);
			fp_rsh(c, b, 1);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("shifting by 2 bits is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			fp_lsh(b, a, 2);
			fp_rsh(c, b, 2);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("shifting by half digit is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			fp_lsh(b, a, FP_DIGIT / 2);
			fp_rsh(c, b, FP_DIGIT / 2);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("shifting by 1 digit is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			fp_lsh(b, a, FP_DIGIT);
			fp_rsh(c, b, FP_DIGIT);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("shifting by 2 digits is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			a[FP_DIGS - 2] = 0;
			fp_lsh(b, a, 2 * FP_DIGIT);
			fp_rsh(c, b, 2 * FP_DIGIT);
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("shifting by 1 digit and half is consistent") {
			fp_rand(a);
			a[FP_DIGS - 1] = 0;
			a[FP_DIGS - 2] = 0;
			fp_lsh(b, a, FP_DIGIT + FP_DIGIT / 2);
			fp_rsh(c, b, (FP_DIGIT + FP_DIGIT / 2));
			TEST_ASSERT(fp_cmp(c, a) == CMP_EQ, end);
		} TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

static int reduction(void) {
	int code = STS_ERR;
	fp_t a, b;
	dv_t t;

	fp_null(a);
	fp_null(b);
	dv_null(t);

	TRY {
		fp_new(a);
		fp_new(b);
		dv_new(t);
		dv_zero(t, 2 * FB_DIGS);

		TEST_BEGIN("modular reduction is correct") {
			fp_rand(a);
			/* Test if a * f(z) mod f(z) == 0. */
			fp_mul(b, a, fp_prime_get());
			TEST_ASSERT(fp_is_zero(b) == 1, end);
		} TEST_END;

#if FP_RDC == BASIC || !defined(STRIP)
		TEST_BEGIN("basic modular reduction is correct") {
			fp_rand(a);
			fp_muln_low(t, a, fp_prime_get());
			fp_rdc_basic(b, t);
			TEST_ASSERT(fp_is_zero(b), end);
		} TEST_END;
#endif

#if FP_RDC == MONTY || !defined(STRIP)
		TEST_BEGIN("montgomery modular reduction is correct") {
			fp_rand(a);
			fp_muln_low(t, a, fp_prime_get());
			fp_rdc_monty(b, t);
			TEST_ASSERT(fp_is_zero(b), end);
		} TEST_END;
#endif

#if FP_RDC == QUICK || !defined(STRIP)
		if (fp_prime_get_spars(NULL) != NULL) {
			TEST_BEGIN("fast modular reduction is correct") {
				fp_rand(a);
				fp_muln_low(t, a, fp_prime_get());
				fp_rdc_quick(b, t);
				TEST_ASSERT(fp_is_zero(b), end);
			}
			TEST_END;
		}
#endif
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fb_free(a);
	fb_free(b);
	dv_free(t);
	return code;
}

static int inversion(void) {
	int code = STS_ERR;
	fp_t a, b, c;

	fp_null(a);
	fp_null(b);
	fp_null(c);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("inversion is correct") {
			fp_rand(a);
			fp_inv(b, a);
			fp_mul(c, a, b);
			fp_set_dig(b, 1);
			TEST_ASSERT(fp_cmp(c, b) == CMP_EQ, end);
		} TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

static int digit(void) {
	int code = STS_ERR;
	fp_t a, b, c, d;
	dig_t g;

	fp_null(a);
	fp_null(b);
	fp_null(c);
	fp_null(d);

	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);
		fp_new(d);

		TEST_BEGIN("addition of a single digit is consistent") {
			fp_rand(a);
			fp_rand(b);
			for (int j = 1; j < FP_DIGS; j++)
				b[j] = 0;
			g = b[0];
			fp_set_dig(b, g);
			fp_add(c, a, b);
			fp_add_dig(d, a, g);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("subtraction of a single digit is consistent") {
			fp_rand(a);
			fp_rand(b);
			for (int j = 1; j < FP_DIGS; j++)
				b[j] = 0;
			g = b[0];
			fp_set_dig(b, g);
			fp_sub(c, a, b);
			fp_sub_dig(d, a, g);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		} TEST_END;

		TEST_BEGIN("multiplication by a single digit is consistent") {
			fp_rand(a);
			fp_rand(b);
			for (int j = 1; j < FP_DIGS; j++)
				b[j] = 0;
			g = b[0];
			fp_set_dig(b, g);
			fp_mul(c, a, b);
			fp_mul_dig(d, a, g);
			TEST_ASSERT(fp_cmp(c, d) == CMP_EQ, end);
		} TEST_END;
	}
	CATCH_ANY {
		ERROR(end);
	}
	code = STS_OK;
  end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	fp_free(d);
	return code;
}

static int divby2(){
	int code = STS_ERR;
	fp_t a, b, c;

        fp_null(a);
	fp_null(b);
	fp_null(c);
	
	TRY {
		fp_new(a);
		fp_new(b);
		fp_new(c);

		TEST_BEGIN("division by 2 is correct") {
			fp_rand(a);
			fp_div2(b, a);
			fp_add(c, b, b);
			TEST_ASSERT(fp_cmp(a, c) == CMP_EQ, end);
		} TEST_END;
	}
	CATCH_ANY {
	}
	code = STS_OK;
end:
	fp_free(a);
	fp_free(b);
	fp_free(c);
	return code;
}

int main(void) {

	core_init();
	
	util_print_banner("Tests for the FP module", 0);

	fp_param_set_any();
	fp_param_print();
	printf("digit size %d \n block size %d \n", FP_DIGIT, FP_DIGS);
	
	util_print_banner("Utilities", 1);
	if (memory() != STS_OK) {
		core_clean();
		return 1;
	}
		printf("two");

	if (util() != STS_OK) {
		core_clean();
		return 1;
	}

	util_print_banner("Arithmetic", 1);
	if (addition() != STS_OK) {
		core_clean();
		return 1;
	}

	if (subtraction() != STS_OK) {
		core_clean();
		return 1;
	}

	if (multiplication() != STS_OK) {
		core_clean();
		return 1;
	}

	if (squaring() != STS_OK) {
		core_clean();
		return 1;
	}

	if (doubling_halving() != STS_OK) {
		core_clean();
		return 1;
	}

	if (shifting() != STS_OK) {
		core_clean();
		return 1;
	}

	if (inversion() != STS_OK) {
		core_clean();
		return 1;
	}

	if (reduction() != STS_OK) {
		core_clean();
		return 1;
	}

	if (digit() != STS_OK) {
		core_clean();
		return 1;
	}

	if (divby2() != STS_OK) {
		core_clean();
		return 1;
	}

	core_clean();
	return 0;
}
